@@ -19,18 +19,13 @@
 #  Foundation, Inc., 59 Temple Place, Suite 330,
 #  Boston, MA 02111-1307, USA.
 
-import gedit
-import gtksourceview2 as gsv
-import gtk
-import glib
 import re
-
+import time
+import xml.sax.saxutils
+from gi.repository import GObject, Pango, PangoCairo, Gdk, Gtk, Gedit
 from signals import Signals
 import constants
-import time
-import pango
-import xml.sax.saxutils
-
+import gettext
 from gpdefs import *
 
 try:
@@ -50,6 +45,7 @@ class DocumentHelper(Signals):
         self._in_mode = False
         self._column_mode = None
         self._move_cursor = None
+        self._previous_move_cursor = None
 
         self._edit_points = []
         self._multi_edited = False
@@ -59,16 +55,22 @@ class DocumentHelper(Signals):
 
         self.connect_signal(self._view, 'notify::buffer', self.on_notify_buffer)
         self.connect_signal(self._view, 'key-press-event', self.on_key_press_event)
-        self.connect_signal(self._view, 'expose-event', self.on_view_expose_event)
+        self.connect_signal(self._view, 'draw', self.on_view_draw)
         self.connect_signal(self._view, 'style-set', self.on_view_style_set)
         self.connect_signal(self._view, 'undo', self.on_view_undo)
         self.connect_signal(self._view, 'copy-clipboard', self.on_copy_clipboard)
         self.connect_signal(self._view, 'cut-clipboard', self.on_cut_clipboard)
         self.connect_signal(self._view, 'paste-clipboard', self.on_paste_clipboard)
         self.connect_signal(self._view, 'query-tooltip', self.on_query_tooltip)
+
         self.connect_signal(self._view, 'move-cursor', self.on_move_cursor)
         self.connect_signal_after(self._view, 'move-cursor', self.on_move_cursor_after)
 
+        try:
+            self.connect_signal(self._view, 'smart-home-end', self.on_smart_home_end)
+        except:
+            pass
+
         self._view.props.has_tooltip = True
 
         self.reset_buffer(self._view.get_buffer())
@@ -86,10 +88,15 @@ class DocumentHelper(Signals):
         return self._in_mode
 
     def _update_selection_tag(self):
-        style = self._view.get_style()
+        context = self._view.get_style_context()
+        state = context.get_state()
 
-        fg = style.text[gtk.STATE_SELECTED]
-        bg = style.base[gtk.STATE_SELECTED]
+        fg_rgba = context.get_color(state)
+        bg_rgba = context.get_background_color(state)
+
+        # TODO: Use GdkRGBA directly when possible
+        fg = Gdk.Color(fg_rgba.red * 65535, fg_rgba.green * 65535, fg_rgba.blue * 65535)
+        bg = Gdk.Color(bg_rgba.red * 65535, bg_rgba.green * 65535, bg_rgba.blue * 65535)
 
         self._selection_tag.props.foreground_gdk = fg
         self._selection_tag.props.background_gdk = bg
@@ -104,7 +111,7 @@ class DocumentHelper(Signals):
 
         self._buffer = None
 
-        if newbuf == None or not isinstance(newbuf, gedit.Document):
+        if newbuf == None or not isinstance(newbuf, Gedit.Document):
             return
 
         if newbuf != None:
@@ -134,27 +141,27 @@ class DocumentHelper(Signals):
         self._view = None
 
         if self._status_timeout != 0:
-            glib.source_remove(self._status_timeout)
+            GObject.source_remove(self._status_timeout)
             self._status_timeout = 0
 
         if self._delete_mode_id != 0:
-            glib.source_remove(self._delete_mode_id)
+            GObject.source_remove(self._delete_mode_id)
             self._delete_mode_id = 0
 
     def initialize_event_handlers(self):
         self._event_handlers = [
             [('Escape',), 0, self.do_escape_mode, True],
             [('Return',), 0, self.do_column_edit, True],
-            [('Return',), gtk.gdk.CONTROL_MASK, self.do_smart_column_edit, True],
-            [('Return',), gtk.gdk.CONTROL_MASK | gtk.gdk.SHIFT_MASK, self.do_smart_column_align, True],
-            [('Return',), gtk.gdk.CONTROL_MASK | gtk.gdk.SHIFT_MASK | gtk.gdk.MOD1_MASK, self.do_smart_column_align, True],
-            [('Home',), gtk.gdk.CONTROL_MASK, self.do_mark_start, True],
-            [('End',), gtk.gdk.CONTROL_MASK, self.do_mark_end, True],
-            [('e',), gtk.gdk.CONTROL_MASK, self.do_toggle_edit_point, True]
+            [('Return',), Gdk.ModifierType.CONTROL_MASK, self.do_smart_column_edit, True],
+            [('Return',), Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK, self.do_smart_column_align, True],
+            [('Return',), Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.MOD1_MASK, self.do_smart_column_align, True],
+            [('Home',), Gdk.ModifierType.CONTROL_MASK, self.do_mark_start, True],
+            [('End',), Gdk.ModifierType.CONTROL_MASK, self.do_mark_end, True],
+            [('e',), Gdk.ModifierType.CONTROL_MASK, self.do_toggle_edit_point, True]
         ]
 
         for handler in self._event_handlers:
-            handler[0] = map(lambda x: gtk.gdk.keyval_from_name(x), handler[0])
+            handler[0] = map(lambda x: Gdk.keyval_from_name(x), handler[0])
 
     def disable_multi_edit(self):
         if self._column_mode:
@@ -162,14 +169,14 @@ class DocumentHelper(Signals):
 
         self._in_mode = False
 
-        self._view.set_border_window_size(gtk.TEXT_WINDOW_TOP, 0)
+        self._view.set_border_window_size(Gtk.TextWindowType.TOP, 0)
         self.remove_edit_points()
 
         if self.toggle_callback:
             self.toggle_callback()
 
     def enable_multi_edit(self):
-        self._view.set_border_window_size(gtk.TEXT_WINDOW_TOP, 20)
+        self._view.set_border_window_size(Gtk.TextWindowType.TOP, 20)
         self._in_mode = True
 
         if self.toggle_callback:
@@ -207,21 +214,7 @@ class DocumentHelper(Signals):
         return True
 
     def iter_to_offset(self, piter):
-        tw = self._view.get_tab_width()
-        start = piter.copy()
-        start.set_line_offset(0)
-
-        offset = 0
-
-        while not start.equal(piter):
-            if start.get_char() == "\t":
-                offset += tw - (offset % tw)
-            else:
-                offset += 1
-
-            start.forward_char()
-
-        return offset
+        return self._view.get_visual_column(piter)
 
     def get_visible_iter(self, line, offset):
         piter = self._buffer.get_iter_at_line(line)
@@ -335,9 +328,15 @@ class DocumentHelper(Signals):
         if not self._in_mode:
             return
 
-        window = self._view.get_window(gtk.TEXT_WINDOW_TOP)
+        window = self._view.get_window(Gtk.TextWindowType.TOP)
         geom = window.get_geometry()
-        window.invalidate_rect(gtk.gdk.Rectangle(0, 0, geom[2], geom[3]), False)
+        #FIXME: there should be some override in pygobject to create the rectangle with values
+        rect = Gdk.Rectangle()
+        rect.x = 0
+        rect.y = 0
+        rect.width = geom[2]
+        rect.height = geom[3]
+        window.invalidate_rect(rect, False)
 
     def _remove_status(self):
         self._status = None
@@ -355,9 +354,9 @@ class DocumentHelper(Signals):
         self._invalidate_status()
 
         if self._status_timeout != 0:
-            glib.source_remove(self._status_timeout)
+            GObject.source_remove(self._status_timeout)
 
-        self._status_timeout = glib.timeout_add(3000, self._remove_status)
+        self._status_timeout = GObject.timeout_add(3000, self._remove_status)
 
     def _apply_column_mode(self):
         mode = self._column_mode
@@ -531,7 +530,7 @@ class DocumentHelper(Signals):
             cp = piter.copy()
             moffset = offset
 
-            if (event.state & gtk.gdk.MOD1_MASK) and cp.backward_visible_cursor_position() and not cp.get_char().isspace():
+            if (event.state & Gdk.ModifierType.MOD1_MASK) and cp.backward_visible_cursor_position() and not cp.get_char().isspace():
                 moffset += 1
 
             if moffset > maxoffset:
@@ -649,7 +648,7 @@ class DocumentHelper(Signals):
 
         return True
 
-    def _draw_column_mode(self, event):
+    def _draw_column_mode(self, cr):
         if not self._column_mode:
             return False
 
@@ -657,15 +656,12 @@ class DocumentHelper(Signals):
         end = self._column_mode[1]
         buf = self._buffer
 
-        col = self._view.get_style().base[gtk.STATE_SELECTED]
         layout = self._view.create_pango_layout('W')
-        width = layout.get_pixel_extents()[1][2]
+        width = layout.get_pixel_extents()[1].width
 
-        ctx = event.window.cairo_create()
-        ctx.rectangle(event.area)
-        ctx.clip()
-
-        ctx.set_source_color(col)
+        context = self._view.get_style_context()
+        col = context.get_background_color(Gtk.StateFlags.SELECTED)
+        Gdk.cairo_set_source_rgba(cr, col)
 
         cstart = self._column_mode[2]
         cend = self._column_mode[3]
@@ -676,14 +672,9 @@ class DocumentHelper(Signals):
             piter = buf.get_iter_at_line(start)
             y, height = self._view.get_line_yrange(piter)
 
-            x_, y = self._view.buffer_to_window_coords(gtk.TEXT_WINDOW_TEXT, 0, y)
+            x_, y = self._view.buffer_to_window_coords(Gtk.TextWindowType.TEXT, 0, y)
             start += 1
 
-            # Check in visible area
-            if y >= event.area.y + event.area.height or \
-               y + height <= event.area.y:
-                continue
-
             # Check where to possible draw fake selection
             start_iter, soff = self.get_visible_iter(start - 1, cstart)
             end_iter, eoff = self.get_visible_iter(start - 1, cend)
@@ -697,8 +688,8 @@ class DocumentHelper(Signals):
             if rw == 0:
                 rw = 1
 
-            ctx.rectangle(rx, y, rw, height)
-            ctx.fill()
+            cr.rectangle(rx, y, rw, height)
+            cr.fill()
 
         return False
 
@@ -770,7 +761,7 @@ class DocumentHelper(Signals):
         return True
 
     def on_key_press_event(self, view, event):
-        defmod = gtk.accelerator_get_default_mod_mask() & event.state
+        defmod = Gtk.accelerator_get_default_mod_mask() & event.state
 
         for handler in self._event_handlers:
             if (not handler[3] or self._in_mode) and event.keyval in handler[0] and (defmod == handler[1]):
@@ -818,6 +809,8 @@ class DocumentHelper(Signals):
 
         if hasattr(where, 'assign'):
             where.assign(iterwas)
+        elif hasattr(Gedit, 'utils_text_iter_assign'):
+            Gedit.utils_text_iter_assign(where, iterwas)
 
         if atinsert:
             buf.move_mark(buf.get_insert(), iterwas)
@@ -875,7 +868,7 @@ class DocumentHelper(Signals):
             # Ooooh, what a hack to be able to work with the undo manager
             self._view.set_editable(False)
             mark = buf.create_mark(None, start, True)
-            self._delete_mode_id = glib.timeout_add(0, self.handle_column_mode_delete, mark)
+            self._delete_mode_id = GObject.timeout_add(0, self.handle_column_mode_delete, mark)
         elif self._edit_points:
             if start.equal(buf.get_iter_at_mark(buf.get_insert())):
                 self.block_signal(buf, 'delete-range')
@@ -976,7 +969,7 @@ class DocumentHelper(Signals):
 
         text = self._column_text()
 
-        clipboard = gtk.Clipboard(self._view.get_display())
+        clipboard = Gtk.Clipboard(self._view.get_display())
         clipboard.set_text(text)
 
         view.stop_emission('copy-clipboard')
@@ -986,7 +979,7 @@ class DocumentHelper(Signals):
             return
 
         text = self._column_text()
-        clipboard = gtk.Clipboard(self._view.get_display())
+        clipboard = Gtk.Clipboard(self._view.get_display())
         clipboard.set_text(text)
 
         view.stop_emission('cut-clipboard')
@@ -1041,7 +1034,7 @@ class DocumentHelper(Signals):
         if not self._edit_points:
             return
 
-        clipboard = gtk.Clipboard(self._view.get_display())
+        clipboard = Gtk.Clipboard(self._view.get_display())
         self._paste_mark = self._buffer.create_mark(None, self._buffer.get_iter_at_mark(self._buffer.get_insert()), True)
 
         clipboard.request_text(self.on_clipboard_text)
@@ -1091,13 +1084,57 @@ class DocumentHelper(Signals):
 
         return piter
 
+    def _move_to_first_char(self, piter, display_line):
+        last = piter.copy()
+
+        if display_line:
+            self._view.forward_display_line_end(last)
+            self._view.backward_display_line_start(piter)
+        else:
+            piter.set_line_offset(0)
+
+            if not last.ends_line():
+                last.forward_to_line_end()
+
+        while piter.compare(last) < 0:
+            if piter.get_char().isspace():
+                piter.forward_visible_cursor_position()
+            else:
+                break
+
+    def _move_to_last_char(self, piter, display_line):
+        first = piter.copy()
+
+        if display_line:
+            self._view.backward_display_line_start(first)
+            self._view.forward_display_line_end(piter)
+        else:
+            if not piter.ends_line():
+                piter.forward_to_line_end()
+
+            first.set_line_offset(0)
+
+        while piter.compare(first) > 0:
+            piter.backward_visible_cursor_position()
+
+            if not piter.get_char().isspace():
+                if not piter.forward_visible_cursor_position():
+                    piter.forward_to_end()
+
+                break
+
+    def _move_edit_point_smart_display_line_ends(self, piter, count):
+        if count > 0:
+            self._move_to_last_char(piter, True)
+        else:
+            self._move_to_first_char(piter, True)
+
+        return piter
+
     def _move_edit_point_display_line_ends(self, piter, count):
         if count > 0:
-            cp = piter.copy()
-
-            if cp.forward_visible_cursor_position() and not self._view.starts_display_line(cp):
-                self._view.forward_display_line_end(piter)
-        elif not self._view.starts_display_line(piter):
+            self._view.forward_display_line_end(piter)
+        else:
             self._view.backward_display_line_start(piter)
 
         return piter
@@ -1110,6 +1147,14 @@ class DocumentHelper(Signals):
 
         return piter
 
+    def _move_edit_point_smart_paragraph_ends(self, piter, count):
+        if count > 0:
+            self._move_to_last_char(piter, False)
+        else:
+            self._move_to_first_char(piter, False)
+
+        return piter
+
     def _move_edit_point_paragraph_ends(self, piter, count):
         if count > 0:
             if not piter.ends_line():
@@ -1124,14 +1169,14 @@ class DocumentHelper(Signals):
 
     def _move_edit_points_by_cursor(self, buf, where):
         actions = {
-            gtk.MOVEMENT_LOGICAL_POSITIONS: self._move_edit_point_logical_positions,
-            gtk.MOVEMENT_VISUAL_POSITIONS: self._move_edit_point_visual_positions,
-            gtk.MOVEMENT_WORDS: self._move_edit_point_words,
-            gtk.MOVEMENT_DISPLAY_LINES: self._move_edit_point_display_lines,
-            gtk.MOVEMENT_DISPLAY_LINE_ENDS: self._move_edit_point_display_line_ends,
-            gtk.MOVEMENT_PARAGRAPHS: self._move_edit_point_paragraphs,
-            gtk.MOVEMENT_PARAGRAPH_ENDS: self._move_edit_point_paragraph_ends,
-            gtk.MOVEMENT_HORIZONTAL_PAGES: self._move_edit_point_horizontal_pages,
+            Gtk.MovementStep.LOGICAL_POSITIONS: self._move_edit_point_logical_positions,
+            Gtk.MovementStep.VISUAL_POSITIONS: self._move_edit_point_visual_positions,
+            Gtk.MovementStep.WORDS: self._move_edit_point_words,
+            Gtk.MovementStep.DISPLAY_LINES: self._move_edit_point_display_lines,
+            Gtk.MovementStep.DISPLAY_LINE_ENDS: self._move_edit_point_display_line_ends,
+            Gtk.MovementStep.PARAGRAPHS: self._move_edit_point_paragraphs,
+            Gtk.MovementStep.PARAGRAPH_ENDS: self._move_edit_point_paragraph_ends,
+            Gtk.MovementStep.HORIZONTAL_PAGES: self._move_edit_point_horizontal_pages,
         }
 
         typ = self._move_cursor[0]
@@ -1147,6 +1192,7 @@ class DocumentHelper(Signals):
             piter = action(piter, count)
             buf.move_mark(point, piter)
 
+        self._previous_move_cursor = self._move_cursor
         self._move_cursor = None
 
     def on_mark_set(self, buf, where, mark):
@@ -1172,7 +1218,7 @@ class DocumentHelper(Signals):
         self.remove_edit_points()
 
     def make_label(self, text, use_markup=True):
-        lbl = gtk.Label(text)
+        lbl = Gtk.Label(text)
         lbl.set_use_markup(use_markup)
         lbl.set_alignment(0, 0.5)
         lbl.show()
@@ -1183,33 +1229,67 @@ class DocumentHelper(Signals):
         if not self._in_mode:
             return False
 
-        geom = view.get_window(gtk.TEXT_WINDOW_TOP).get_geometry()
+        geom = view.get_window(Gtk.TextWindowType.TOP).get_geometry()
 
         if x < geom[0] or x > geom[0] + geom[2] or y < geom[1] or y > geom[1] + geom[3]:
             return False
 
-        table = gtk.Table(12, 2)
+        table = Gtk.Table(13, 2)
         table.set_row_spacings(3)
         table.set_col_spacings(12)
 
-        table.attach(self.make_label('<b>Selection</b>', True), 0, 2, 0, 1, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
+        table.attach(self.make_label('<b>Selection</b>', True),
+                     0, 2, 0, 1,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<Enter>', False),
+                     0, 1, 1, 2,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<Ctrl><Enter>', False),
+                     0, 1, 2, 3,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<Ctrl><Shift><Enter>', False),
+                     0, 1, 3, 4,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<Ctrl><Alt><Shift><Enter>', False),
+                     0, 1, 4, 5,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
 
-        table.attach(self.make_label('<Enter>', False), 0, 1, 1, 2, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-        table.attach(self.make_label('<Ctrl><Enter>', False), 0, 1, 2, 3, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-        table.attach(self.make_label('<Ctrl><Shift><Enter>', False), 0, 1, 3, 4, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-        table.attach(self.make_label('<Ctrl><Alt><Shift><Enter>', False), 0, 1, 4, 5, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-
-        sep = gtk.HSeparator()
+        sep = Gtk.HSeparator()
         sep.show()
 
-        table.attach(sep, 0, 2, 5, 6, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-
-        table.attach(self.make_label('<b>Edit points</b>', True), 0, 2, 6, 7, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-        table.attach(self.make_label('<Ctrl>+E', False), 0, 1, 7, 8, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-        table.attach(self.make_label('<Ctrl><Home>', False), 0, 1, 8, 9, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-        table.attach(self.make_label('<Ctrl><End>', False), 0, 1, 9, 10, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-        table.attach(self.make_label('<Ctrl><Shift><Enter>', False), 0, 1, 10, 11, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
-        table.attach(self.make_label('<Ctrl><Alt><Shift><Enter>', False), 0, 1, 10, 11, gtk.SHRINK | gtk.FILL, gtk.SHRINK | gtk.FILL)
+        table.attach(sep,
+                     0, 2, 5, 6,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<b>Edit points</b>', True),
+                     0, 2, 6, 7,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<Ctrl>+E', False),
+                     0, 1, 7, 8,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<Ctrl><Home>', False),
+                     0, 1, 8, 9,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<Ctrl><End>', False),
+                     0, 1, 9, 10,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<Ctrl><Shift><Enter>', False),
+                     0, 1, 10, 11,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
+        table.attach(self.make_label('<Ctrl><Alt><Shift><Enter>', False),
+                     0, 1, 11, 12,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL,
+                     Gtk.AttachOptions.SHRINK | Gtk.AttachOptions.FILL)
 
         table.attach(self.make_label(_('Enter column edit mode using selection')), 1, 2, 1, 2)
         table.attach(self.make_label(_('Enter <b>smart</b> column edit mode using selection')), 1, 2, 2, 3)
@@ -1226,56 +1306,59 @@ class DocumentHelper(Signals):
         tooltip.set_custom(table)
         return True
 
-    def from_color(self, col):
-        return [col.red / float(0x10000), col.green / float(0x10000), col.blue / float(0x10000)]
-
     def _background_color(self):
-        col = self.from_color(self._view.get_style().base[self._view.state])
-        if col[2] > 0.8:
-            col[2] -= 0.2
+        context = self._view.get_style_context()
+        col = context.get_background_color(context.get_state())
+        if col.green > 0.8:
+            col.green -= 0.2
         else:
-            col[2] += 0.2
+            col.green += 0.2
 
         return col
 
-    def on_view_expose_event(self, view, event):
-        if event.window == view.get_window(gtk.TEXT_WINDOW_TEXT):
-            return self._draw_column_mode(event)
+    def on_view_draw(self, view, cr):
+        window = view.get_window(Gtk.TextWindowType.TEXT)
+        if Gtk.cairo_should_draw_window (cr, window):
+            return self._draw_column_mode(cr)
 
-        if event.window != view.get_window(gtk.TEXT_WINDOW_TOP):
+        window = view.get_window(Gtk.TextWindowType.TOP)
+        if window is None or not Gtk.cairo_should_draw_window (cr, window):
             return False
 
         if not self._in_mode:
             return False
 
-        ctx = event.window.cairo_create()
         col = self._background_color()
 
-        ctx.set_source_rgb(col[0], col[1], col[2])
-        ctx.rectangle(event.area)
-        ctx.fill_preserve()
-        ctx.clip()
+        Gdk.cairo_set_source_rgba(cr, col)
+        cr.fill_preserve()
 
         layout = view.create_pango_layout(_('Multi Edit Mode'))
 
-        layout.set_font_description(pango.FontDescription('Sans 10'))
+        layout.set_font_description(Pango.FontDescription('Sans 10'))
         extents = layout.get_pixel_extents()
 
-        geom = event.window.get_geometry()
+        w = window.get_width()
+        h = window.get_height()
 
-        ctx.translate(0.5, 0.5)
-        ctx.set_line_width(1)
+        Gtk.cairo_transform_to_window (cr, view, window)
 
-        col = self.from_color(view.get_style().text[view.state])
+        cr.translate(0.5, 0.5)
+        cr.set_line_width(1)
 
-        ctx.set_source_rgba(col[0], col[1], col[2], 0.6)
-        ctx.move_to(geom[0], geom[1] + geom[3] - 1)
-        ctx.rel_line_to(geom[2], 0)
-        ctx.stroke()
+        context = view.get_style_context()
+        col = context.get_color(context.get_state())
+        col.alpha = 0.6
+        Gdk.cairo_set_source_rgba(cr, col)
 
-        ctx.set_source_rgb(col[0], col[1], col[2])
-        ctx.move_to(geom[2] - extents[1][2] - 3, (geom[3] - extents[1][3]) / 2)
-        ctx.show_layout(layout)
+        cr.move_to(0, h - 1)
+        cr.rel_line_to(w, 0)
+        cr.stroke()
+
+        col.alpha = 1.0
+        Gdk.cairo_set_source_rgba(cr, col)
+        cr.move_to(w - extents[1].width - 3, (h - extents[1].height) / 2)
+        PangoCairo.show_layout(cr, layout)
 
         if not self._status:
             status = ''
@@ -1283,17 +1366,37 @@ class DocumentHelper(Signals):
             status = str(self._status)
 
         if status:
-            layout.set_markup(status)
+            layout.set_markup(status, -1)
 
-            ctx.move_to(3, (geom[3] - extents[1][3]) / 2)
-            ctx.show_layout(layout)
+            cr.move_to(3, (h - extents[1].height) / 2)
+            PangoCairo.show_layout(cr, layout)
 
         return False
 
+    def on_smart_home_end(self, view, iter, count):
+        if not self._in_mode or not self._edit_points or not self._previous_move_cursor:
+            return
+
+        ins = self._buffer.get_iter_at_mark(self._buffer.get_insert())
+
+        if not ins.equal(iter):
+            return
+
+        if self._previous_move_cursor[0] == Gtk.MovementStep.DISPLAY_LINE_ENDS:
+            cb = self._move_edit_point_smart_display_line_ends
+        else:
+            cb = self._move_edit_point_smart_paragraph_ends
+
+        for point in self._edit_points:
+            piter = self._buffer.get_iter_at_mark(point)
+            piter = cb(piter, count)
+            self._buffer.move_mark(point, piter)
+
     def on_move_cursor(self, view, step_size, count, extend_selection):
         self._move_cursor = [step_size, count]
 
     def on_move_cursor_after(self, view, step_size, count, extend_selection):
         self._move_cursor = None
+        self._previous_move_cursor = None
 
 # ex:ts=4:et:
